
/*****************************************************************************
	old win_count with tilted time windows
	(From term paper)
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#define MAXINT INT_MAX

#include "hcwin.h"

extern long N;		// current transaction number
extern long BSIZE;	// batch size

/***************************************************************************** 
win_type *win_init(double rho, double epsilon, long M)
INPUTS:
	rho - hCount parameter
	epsilon - hCount parameter
	M - largest item in the stream

OUTPUT:
	Pointer to initialized win_type struct

Creates an instance of the win_type struct and initializes everything
*****************************************************************************/
win_type *win_init(double rho, double epsilon, long M)
{
    double e;			// min x/ln(x) and x > 1  --> evaluates to the base of the natural logarithm
    int R_RANGE;		// range for random numbers a_i and b_i
   
	int i, j, k;
    S_TYPE *aptr;

    win_type *result = calloc(1, sizeof(win_type));
    if (result == NULL) {
		printf("\ncalloc error");
		exit(100);
    }

    result->N = 0;			// total transactions
    result->M = M; 
    
	e = 2.718281828;
    //result->P = 1048583;	// big prime number
	//result->P = 1021;	
	result->P = 131101;

    R_RANGE = result->M;
    
	result->m =  e / epsilon;		// length of hash table 
    result->h = (int) (log(-result->M / log(rho)));		// width of hash table
    
	// Allocate memory for the counters S
	aptr = (S_TYPE *) malloc(result->m * result->h * sizeof(S_TYPE));
    if (aptr == NULL) {
		printf("\nmalloc() error.");
		exit(100);
    }
    
	result->S = (S_TYPE **) malloc(result->m *sizeof(S_TYPE));
    if (result->S == NULL) {
		printf("\nmalloc() error.");
		exit(200);
    }
    for (i = 0; i < result->m; i++) {
		result->S[i] = aptr + i * result->h;
    }
    
	// Set all counters to 0
	memset((void *) aptr, 0, result->m * result->h * sizeof(S_TYPE) / sizeof(char));
    
	// generate h pairs of random numbers a_i and b_i
	result->a = (int *) malloc(result->h * sizeof(int));
    if (result->a == NULL) {
		printf("\nmalloc() error.");
		exit(300);
    }
    
	result->b = (int *) malloc(result->h * sizeof(int));
    if (result->b == NULL) {
		printf("\nmalloc() error.");
		exit(400);
    }
    for (i = 0; i < result->h; i++) {
		result->a[i] = random(0, R_RANGE);
		result->b[i] = random(0, R_RANGE);
    }

	/* mark empty intermediate windows by -1 */
	for (i=0; i<result->m; i++) {
		for (j=0; j<result->h; j++) {
			for (k=2; k<=WINDOWS*2+1; k+=2) {
				result->S[i][j].c[k] = -1;
			}
		}
	}
	return (result);
}

/*****************************************************************************
void win_update(win_type * hc, int k, int ttype)
INPUTS:
	hc - pointer to the hCount struct
	k - the item in the stream
	ttype - transaction type: 1 for INSERT, -1 for DELETE

OUTPUT:
	NIL

Updates the hCount table with new transaction. 
*****************************************************************************/
void win_update(win_type *hc, int val, int ttype)
{
	int pos, i, j, k;
	int carry, tmp;

	N++;
	/* if (ttype == INSERT) {
		N++;
	} else {
		N--;
	} */

	/* shift once every BSIZE transactions */
	if (N % BSIZE == 0) {
		for (j=0; j < hc->m; j++) {
			for (k=0; k < hc->h; k++) {
				carry = hc->S[j][k].c[0];
				hc->S[j][k].c[0] = 0;
				i = 1;

				while (1) {
					tmp = hc->S[j][k].c[i];
					hc->S[j][k].c[i] = carry;
					carry = tmp;

					if (hc->S[j][k].c[i+1] == -1) {
						hc->S[j][k].c[i+1] = carry;
						break;
					} else {
						carry += hc->S[j][k].c[i+1];
						hc->S[j][k].c[i+1] = -1;
						i+=2;
					}
				}
			}
		}
  	}

	/* Update the 0 level counters */
	for (j=0; j < hc->h; j++) {
		pos = ((((long long) hc->a[j] * val + hc->b[j]) % hc->P) % hc->m);
		if (ttype == INSERT) {
			hc->S[pos][j].c[0]++;
		} else {
			hc->S[pos][j].c[0]--;
		}
	}
}

/*****************************************************************************
long win_query(win_type *hc, int id, long item, double alpha)
INPUTS:
	hc - pointer to the hCount struct
	val - item to look for
	window - 
	alpha - 

OUTPUT:
	Count of "item" 

Counts the number of occurences of "item" in window number "id" (This is the
core query part of the hCount algorithm)
*****************************************************************************/
long win_query(win_type *hc, long val, long window, double alpha)
{
	int min, pos, c, j;
	int idx, depth;
	long win_done = 0;
	int cnext = 0;

	min = MAXINT;
	for (j=0; j < hc->h; j++) {
		depth = -1;
		idx = 0;
		pos = ((((long long) hc->a[j] * val + hc->b[j]) % hc->P) % hc->m);
		c = hc->S[pos][j].c[0];
        win_done = BSIZE;

		if(window <= BSIZE) {
			c = 0;
        } else {
    		while (win_done < window) {
                depth++;
    			win_done += BSIZE * (1 << depth);
    			if(win_done > window)
    			    break;
                idx++;

    			if (hc->S[pos][j].c[idx] != -1) {
    		    	c += hc->S[pos][j].c[idx];
    			}
                idx++;
    		}
    	}

        cnext = c;
        cnext += hc->S[pos][j].c[idx +1];

		/* do the weighted average */
		c = c + alpha * (cnext - c);

		if (c < min) {
			min = c;
		}
	}

	return(min);
}

/*****************************************************************************
int win_size(win_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	Space occupied by the hCount structure in bytes
*****************************************************************************/
int win_size(win_type * hc)
{
    int size;
    size = hc->m * hc->h * sizeof(S_TYPE);
    return (size);
}


/*****************************************************************************
void win_print(win_type *hc, int id)
INPUTS:
	hc - pointer to the hCount structure
	id - ID of Tilted Time Window

OUTPUT:
	NIL

Prints relevant debug information about the hCount structure
*****************************************************************************/
void win_print(win_type *hc, int id)
{
    int i, j;

    for (i = 0; i < hc->m; i++) {
		printf("\n");
		for (j = 0; j < hc->h; j++)
			printf("%d ", hc->S[i][j].c[id]);
    }

    for (i = 0; i < hc->h; i++) {
		printf("\n%d, a=%d, b=%d", i, hc->a[i], hc->b[i]);
    }

    printf("\nm = %d", hc->m);
    printf("\nh = %d", hc->h);
    printf("\nP = %d", hc->P);
    printf("\nwin_Print -- END");

}

/*****************************************************************************
void win_destroy(win_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	NIL

Frees up memory used by the hCount structure
*****************************************************************************/
void win_destroy(win_type * hc)
{
    free(hc->a);
    free(hc->b);
    free(hc->S);
}

