
/*****************************************************************************
	TiTiCount with subtraction algorithm. The query part
	01-NOV-2007
*****************************************************************************/

#include <stdio.h>
#include <math.h>

#include <assert.h>
#include "sub.h"
#include "hcsub.h"


#include <limits.h>
#define MAXINT INT_MAX

#define	FALSE	0
#define	TRUE	1

extern long N;		// current transaction count
extern long BSIZE;		// batch size

long num_transactions(int id);



/*****************************************************************************
long query(hc_type *hc, long tqmin, long tqmax, long ITEM)
INPUTS:
	hc - the hCount structure
	tqmin - lower transaction number
	tqmax - upper transaction number
	ITEM - the item to find

OUTPUT:
	Number of occurences of ITEM in the range (tqmin, tqmax) of transactions
*****************************************************************************/
long query(hc_type *hc, long tqmin, long tqmax, long ITEM)
{
    long cnt;
    int im_small;					// immediate smaller window (see Base Case)
    int ws;
    long twsmin;
    double frac, qdiff;

    check_query(tqmin, tqmax);		// check validity of query
    qdiff = (tqmax-tqmin);

    ws = smallest_window(tqmax);	// find smallest window containing tqmax
    twsmin = min_boundary(ws);		// find starting transaction number in ws
    //	printf("\ntqmax = %ld, smallest_window = %d, min_boundary = %ld", tqmax, ws, twsmin);

    /****************** BASE CASE **************************

						 tqmin  tqmax
    OLD                    |      |                     NEW
	----------|----------|-{------}-|----------|----------|
                                                 |_____________________| <-- ws
						 |          
						 |          |__________| <-- immediate smaller overlapping window
					   twsmin		

	********************************************************/

    if (twsmin <= tqmin) {	// base case
        //printf("\nSUBQUERY: %ld, %ld", tqmin, tqmax);

        im_small = check_overlapping(ws);	// check if there is smaller overlapping window
        if (im_small != -1) {
            frac = qdiff / (num_transactions(ws) - num_transactions(im_small));
            cnt = frac * count_with_subtraction(hc, ws, im_small, ITEM);
            //printf("\nfrac = %f", frac);
            //printf("-- SUBTRACTION NEEDED. \nwin%d - win%d. COUNT: %ld", ws, im_small, cnt);
        } else {
            frac = qdiff / num_transactions(ws);
            cnt = frac * count_single_window(hc, ws, ITEM);
            //printf("\nfrac = %f", frac);
            //printf("-- NO SUBTRACTION. win%d. COUNT: %ld", ws, cnt);
        }
        return(cnt);
    } else {	// split into smaller windows recursively

        return(query(hc, twsmin, tqmax, ITEM) + query(hc, tqmin, twsmin-1, ITEM));

    }

}

long num_transactions(int id)
{
    if (id==0) {
        id = 1;
    }
    return((long)(BSIZE * pow(2.0, id-1)));
}

/*****************************************************************************
int smallest_window(long tran)
INPUTS:
	tran - transaction number

OUTPUT:
	The ID of the smallest window which contains the transaction tran
*****************************************************************************/
int smallest_window(long tran)
{
    /* This list will store the starting transaction number for each window.
		Assume 20 windows for the moment, tidy up later */
    long wfirst[20];		 // xxx will work only if there are 20 windows.
    int cnt=0;

    get_window_firsts(wfirst);

    while (wfirst[cnt] > tran) {
        cnt++;
    }

    return(cnt);
}


/*****************************************************************************
void get_window_firsts(long *wfirst)
INPUTS:
	wfirst - array to store the first transaction in each window
	
	uses N, the global current transaction count
	uses BSIZE, the global batch size

OUTPUT:
	NIL. 

Generates the wfirst array.
*****************************************************************************/
void get_window_firsts(long *wfirst)
{
    int k;
    int num;	// number of windows in use
    num = (int)((double)log((N-1)/BSIZE)/log(2.0)+1);		// log to the base 2
    //printf("\nNo. of windows: %d", num);

    for (k=0; k<=num; k++) {
        wfirst[k] = min_boundary(k);
        //printf("\nwfirst[%d] = %ld", k, wfirst[k]);
    }
    //getchar();
}


/*****************************************************************************
long min_boundary(int wID);
INPUTS:
	wID - window ID

	uses N, the global current transaction count
	uses BSIZE, the global batch size

OUTPUT:
	Smallest transaction number stored in window 'wID' (which is also the 
	min boundary of wID)
*****************************************************************************/
long min_boundary(int wID)
{
    int p, q;

    p = (N-1)/BSIZE;

    if (wID == 0) {		// current window - special case
        return(p * BSIZE + 1); // that, is N. wrong?
    } else {
        q = p % (long)pow(2, wID-1);
        return((p-q-pow(2,wID-1)) * BSIZE + 1);
    }
}


/*****************************************************************************
int check_overlapping(int wID)
INPUTS:
	wID - window ID

	uses BSIZE, the global batch size

OUTPUT:
	-1 if there is no smaller window overlapping with wID
	ELSE,
	ID of the immediate smaller window overlapping with wID
*****************************************************************************/
int check_overlapping(int wID)
{
    int FOUND = FALSE;
    int i;
    long last;	// last transaction in window wID

    last = min_boundary(wID) + pow(2, wID-1) * BSIZE - 1;

    for (i=wID-1; i>0; i--) {
        if (min_boundary(i) < last) {
            FOUND = TRUE;
            break;
        }
    }

    if (FOUND) {
        return(i);
    } else {
        return(-1);
    }
}


/*****************************************************************************
long count_single_window(hc_type *hc, int wID, long ITEM)
INPUTS:
	hc - the hCount structure
	wID - ID of window
	ITEM - item to count

OUTPUT:
	count of ITEM in wID
*****************************************************************************/
long count_single_window(hc_type *hc, int wID, long ITEM)
{
    return(hc_query(hc, wID, ITEM));
}


/*****************************************************************************
long count_with_subtraction(hc_type *hc, int wID1, int wID2, long ITEM)
INPUTS:
	hc - the hCount structure
	wID1 - ID of larger window
	wID2 - ID of smaller window (overlapping with wID1)
	ITEM - item to count

OUTPUT:
	count of ITEM in wID1 minus count of ITEM in wID2
*****************************************************************************/
long count_with_subtraction(hc_type *hc, int wID1, int wID2, long ITEM)
{
    return(count_single_window(hc, wID1, ITEM) - count_single_window(hc, wID2, ITEM));
}


/*****************************************************************************
void check_query(long tqmin, long tqmax)
INPUTS:
	tqmin - lower transaction number
	tqmax - upper transaction number

OUTPUT:
	NIL

Makes sure that the query is valid and that we can answer the query
based on the history stored so far. If not, print error message and quit.
*****************************************************************************/
void check_query(long tqmin, long tqmax)
{

}
